@@ -1,5 +1,5 @@
 /*
- *	$Id: ipfwadm.c,v 1.13 1996/05/24 13:36:55 jos Exp $
+ *	$Id: ipfwadm.c,v 1.14 1996/06/17 10:28:52 jos Exp $
  *
  *
  *	ipfwadm -- IP firewall and accounting administration
@@ -76,9 +76,16 @@
  *	1.13	Also support kernels 1.3.66 up to 1.99.6 (untested).
  *		Check on missing command.
  *		[ipfwadm 2.1]
+ *	1.14	Add <errno.h> to make new compile environments happy.
+ *		Allow -r without a port number (use 0 as default value).
+ *		Add -s command for setting timeout values (masquerading).
+ *		[ipfwadm 2.2]
+ *		
  */
 
 #include <stdio.h>
+#include <errno.h>
+#include <limits.h>
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -91,18 +98,24 @@
 #include <linux/icmp.h>
 #include <linux/if.h>
 #include <linux/ip_fw.h>
+#include <asm/param.h>
 
 #ifndef	IP_FW_F_REDIR
 #define IPFWADM_NO_REDIR
-#define IP_FW_F_REDIR	0
+#define IP_FW_F_REDIR		0
 #endif	/* ! IP_FW_F_REDIR */
 
 #if	!defined(IP_FW_F_ACCTIN) || !defined(IP_FW_F_ACCTOUT)
 #define IPFWADM_NO_ACCTDIR
-#define IP_FW_F_ACCTIN	0
-#define IP_FW_F_ACCTOUT	0
+#define IP_FW_F_ACCTIN		0
+#define IP_FW_F_ACCTOUT		0
 #endif	/* ! IP_FW_F_REDIR */
 
+#ifndef	IP_FW_MASQ_TIMEOUTS
+#define IPFWADM_NO_TIMEOUT
+#define IP_FW_MASQ_TIMEOUTS	0
+#endif	/* ! IP_FW_MASQ_TIMEOUTS */
+
 #define IP_VERSION	4
 #define IP_OFFSET	0x1FFF
 
@@ -123,6 +136,7 @@
 #define CMD_CHECK	0x0040
 #define CMD_HELP	0x0080
 #define CMD_INSERT	0x0100
+#define CMD_TIMEOUT	0x0200
 
 #define OPT_NONE	0x000000
 #define OPT_EXTENDED	0x000001
@@ -166,8 +180,14 @@ struct masq {
 	short		pdelta;		/* Delta in sequence numbers before last */
 };
 
-char ipfwadm_version[] = "$Id: ipfwadm.c,v 1.13 1996/05/24 13:36:55 jos Exp $";
-char package_version[] = "ipfwadm 2.1, 1996/05/24";
+struct masq_timeout {
+	int	tcp_timeout;
+	int	tcp_fin_timeout;
+	int	udp_timeout;
+} timeouts;
+
+char ipfwadm_version[] = "$Id: ipfwadm.c,v 1.14 1996/06/17 10:28:52 jos Exp $";
+char package_version[] = "ipfwadm 2.2, 1996/06/17";
 
 int ssocmd_insert[4] =
 	{IP_FW_INSERT_FWD, IP_FW_INSERT_IN, IP_FW_INSERT_OUT, IP_ACCT_INSERT};
@@ -213,6 +233,7 @@ extern void parse_all_ports(char **, unsigned short *, int, int);
 extern unsigned short parse_port(char *, unsigned short);
 extern void store_port(char *, unsigned short *, int, char *[]);
 extern void parse_hexbyte(char *, unsigned char *);
+extern int parse_timeout(char *);
 
 extern struct in_addr *host_to_addr(char *, int *);
 extern char *addr_to_host(struct in_addr *);
@@ -224,7 +245,7 @@ extern char *addr_to_dotted(struct in_addr *);
 extern char *mask_to_dotted(struct in_addr *);
 extern int service_to_port(char *, unsigned short);
 extern char *port_to_service(int, unsigned short);
-extern int string_to_number(char *, int, int);
+extern long string_to_number(char *, long, long);
 extern char *policy_to_string(int);
 
 extern int add_delete_entries(int, int);
@@ -255,15 +276,14 @@ main(int argc, char *argv[])
 
 	program = argv[0];
 
-	while ((c = getopt(argc, argv, "AFIMOadilzfp:chP:S:D:V:W:bekmnor:t:vxy")) != -1)
+	while ((c = getopt(argc, argv, "AFIMOadilzfs:p:chP:S:D:V:W:bekmnort:vxy")) != -1)
 		switch (c) {
 		case 'A':
 			if (chain != CHN_NONE)
 				exit_error(2, "multiple categories specified");
 			chain = CHN_ACCT;
-			if (optind < argc && argv[optind][0] != '-') {
+			if (optind < argc && argv[optind][0] != '-')
 				parse_direction(argv[optind++]);
-			}
 			break;
 		case 'F':
 			if (chain != CHN_NONE)
@@ -333,6 +353,25 @@ main(int argc, char *argv[])
 			command = CMD_POLICY;
 			parse_policy(optarg);
 			break;
+		case 's':
+#ifndef	IPFWADM_NO_TIMEOUT
+			if (command != CMD_NONE)
+				exit_error(2, "multiple commands specified");
+			command = CMD_TIMEOUT;
+			if (optind + 1 < argc && argv[optind][0] != '-' &&
+					argv[optind+1][0] != '-') {
+				timeouts.tcp_timeout =
+					HZ * parse_timeout(optarg);
+				timeouts.tcp_fin_timeout =
+					HZ * parse_timeout(argv[optind++]);
+				timeouts.udp_timeout =
+					HZ * parse_timeout(argv[optind++]);
+			} else
+				exit_error(2, "-s requires 3 timeout values");
+#else	/* IPFWADM_NO_TIMEOUT */
+			exit_error(2, "setting masquerading timeouts not supported");
+#endif	/* IPFWADM_NO_TIMEOUT */
+			break;
 		case 'c':
 			if (command != CMD_NONE)
 				exit_error(2, "multiple commands specified");
@@ -394,7 +433,10 @@ main(int argc, char *argv[])
 #ifndef	IPFWADM_NO_REDIR
 			check_option(OPT_REDIRECT, 'r');
 			firewall.fw_flg |= IP_FW_F_REDIR;
-			rport = optarg;
+			if (optind < argc && argv[optind][0] != '-')
+				rport = argv[optind++];
+			else
+				rport = "0";
 #else	/* IPFWADM_NO_REDIR */
 			exit_error(2, "redirection not supported");
 #endif	/* IPFWADM_NO_REDIR */
@@ -447,8 +489,10 @@ main(int argc, char *argv[])
 			options & OPT_POLICY)
 		/* command is CMD_INSERT, CMD_APPEND, or CMD_DELETE */
 		exit_error(2, "no policy allowed with non-firewall entries");
-	else if (chain == CHN_MASQ && !(command & CMD_LIST))
-		exit_error(2, "-M only allowed in combination with -l command");
+	else if (chain == CHN_MASQ && !(command & CMD_TIMEOUT) && !(command & CMD_LIST))
+		exit_error(2, "-M only allowed in combination with -s or -l command");
+	else if (chain != CHN_MASQ && command & CMD_TIMEOUT)
+		exit_error(2, "setting masquerading timeouts requires -M flag");
 
 	if ((options & OPT_BIDIR) &&
 			!(command & (CMD_INSERT | CMD_APPEND | CMD_DELETE)))
@@ -564,6 +608,9 @@ main(int argc, char *argv[])
 	case CMD_RESET:
 		ret = do_setsockopt(ssocmd_zero[chain], &dummy, sizeof(dummy));
 		break;
+	case CMD_TIMEOUT:
+		ret = do_setsockopt(IP_FW_MASQ_TIMEOUTS, &timeouts, sizeof(timeouts));
+		break;
 	case CMD_LIST:
 	case CMD_LIST | CMD_RESET:
 		if (chain == CHN_MASQ)
@@ -875,6 +922,22 @@ parse_hexbyte(char *s, unsigned char *b)
 	}
 }
 
+int
+parse_timeout(char *s)
+{
+	int timeout;
+
+	if ((timeout = string_to_number(s, 0, INT_MAX)) != -1)
+		return timeout;
+	else {
+		fprintf(stderr, "%s: invalid timeout value \"%s\" specified\n",
+			program, s);
+		exit_tryhelp(2);
+		/* make the compiler happy... */
+		return 0;
+	}
+}
+
 struct in_addr *
 host_to_addr(char *name, int *naddr)
 {
@@ -1050,8 +1113,8 @@ port_to_service(int port, unsigned short kind)
 		return (char *) NULL;
 }
 
-int
-string_to_number(char *s, int min, int max)
+long
+string_to_number(char *s, long min, long max)
 {
 	long number;
 	char *end;
@@ -1060,7 +1123,7 @@ string_to_number(char *s, int min, int max)
 	if (*end == '\0' && end != s) {
 		/* we parsed a number, let's see if we want this */
 		if (min <= number && number <= max)
-			return (int) number;
+			return number;
 		else
 			return -1;
 	} else
@@ -1445,7 +1508,9 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 
 	if (flags & IP_FW_F_REDIR) {
 		i = fw->fw_nsp + fw->fw_ndp;
-		if (format & FMT_NUMERIC)
+		if (!fw->fw_pts[i])
+			fputs((format & FMT_NUMERIC) ? " => *" : " => any", fp);
+		else if (format & FMT_NUMERIC)
 			fprintf(fp, " => %u", fw->fw_pts[i]);
 		else if ((service = port_to_service(fw->fw_pts[i], kind)) != NULL)
 			fprintf(fp, " => %s", service);
@@ -1746,7 +1811,7 @@ exit_printhelp()
 		"       %s -F command [options] (forwarding firewall)\n"
 		"       %s -I command [options] (input firewall)\n"
 		"       %s -O command [options] (output firewall)\n"
-		"       %s -M -l [options] (masquerading entries)\n"
+		"       %s -M [-s | -l] [options] (masquerading entries)\n"
 		"       %s -h (print this help information))\n\n",
 		package_version, program, program, program, program,
 		program, program);
@@ -1759,6 +1824,8 @@ exit_printhelp()
 		"  -z		reset packet/byte counters of all rules of this category\n"
 		"  -f		remove all rules of this category\n"
 		"  -p policy	change default policy (accept/deny/reject)\n"
+		"  -s tcp tcpfin udp\n"
+		"		set masuerading timeout values\n"
 		"  -c		check acceptance of IP packet\n\n"
 		"Options:\n"
 		"  -P		protocol (either tcp, udp, icmp, or all)\n"
@@ -1774,7 +1841,7 @@ exit_printhelp()
 		"  -m		masquerade packets as coming from local host\n"
 		"  -n		numeric output of addresses and ports\n"
 		"  -o		turn on kernel logging for matching packets\n"
-		"  -r port	redirect packets to local port (transparent proxying)\n"
+		"  -r [port]	redirect packets to local port (transparent proxying)\n"
 		"  -t and xor	and/xor masks for TOS field\n"
 		"  -v		verbose mode\n"
 		"  -x		expand numbers (display exact values)\n"
